/*

Copyright (c) 2009 Şükrü Çınar  (sukru cinar)

This program is distributed under GNU GPL v2. See file COPYING
for licensing details.

*/

#include <stdlib.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "parsedoc.h"
#include "style.h"
#include "miscfn.h"

#include "decl.h"

static void parse_declaration (declist_t * L, grid_t * G, int i, int j);
static char *get_cell (grid_t * G, int r, int c);

void *__unused_decl_c[] = {

  0,
};

void declist_init (declist_t * L)
{
  L->head.prev = L->head.next = &L->head;
}

decl_t *declist_destroy (declist_t * L)
{
  typeof (L->head.next) N;
  decl_t *R;

  N = L->head.next;
  if (N == &L->head)
    return 0;
  R = N->data;

  declist_remove (L, N);
  return R;
}

declist_t *declist_new ()
{
  declist_t *L;
  L = (typeof (L)) malloc (sizeof (*L));
  L->head.prev = L->head.next = &L->head;
  return L;
}

void declist_append (declist_t * L, decl_t * data)
{
  struct declist_t_NODE *n;
  n = (typeof (n)) malloc (sizeof (n[0]));
  n->data = data;
  n->prev = L->head.prev;
  n->next = &L->head;
  L->head.prev->next = n;
  L->head.prev = n;
}

void declist_prepend (declist_t * L, decl_t * data)
{
  struct declist_t_NODE *n;
  n = (typeof (n)) malloc (sizeof (n[0]));
  n->data = data;
  n->next = L->head.next;
  n->prev = &L->head;
  L->head.next->prev = n;
  L->head.next = n;
}

void declist_remove (declist_t * L, struct declist_t_NODE *n)
{
  n->prev->next = n->next;
  n->next->prev = n->prev;
  free (n);
}

int declist_isempty (declist_t * L)
{
  return L->head.next == &L->head;
}

declist_t *
parse_declarations (grid_t * G)
{
  declist_t *L;
  int i, j;
  L = declist_new ();
  for (i = 0; i < G->nrows; i++)
    for (j = 0; j < G->ncols; j++)
      if (G->cells[i][j].kind == ck_data)
	{
	  parse_declaration (L, G, i, j);
	  j = G->ncols;
	}
  return L;

}

static void parse_declaration (declist_t * L, grid_t * G, int i, int j)
{
  decl_t *D;
  char *C;

  D = (typeof (D)) calloc (1, sizeof (D[0]));

  D->name = get_cell (G, i, j);
  if (!D->name)
    exit (fprintf (stderr, "empty name in "
		   "declaration at row %d\n", i + 1));
  j++;
  C = get_cell (G, i, j);

  if (!C)
    exit (fprintf (stderr, "empty type in "
		   "declaration at row %d\n", i + 1));
  {
    int _cmp10;
    int _swvar10;
    char *_strvar10;

    enum case_values
    {
      __UNKNOWN_10,

      STRSW_int_10,

      STRSW_str_10,

    };

    _strvar10 = C;

    _swvar10 = __UNKNOWN_10;

    _cmp10 = strcmp (_strvar10, "int");
    if (_cmp10 > 0)
      {
	_cmp10 = strcmp (_strvar10, "str");
	if (_cmp10 == 0)
	  {
	    _swvar10 = STRSW_str_10;
	  }
      }
    else if (_cmp10 == 0)
      {
	_swvar10 = STRSW_int_10;
      }

    switch (_swvar10)
      {
/* I remove this because the default case doesnt work otherwise..
		case __UNKNOWN_10:
			break;
*/

      case STRSW_int_10:

	D->type = typ_int;
	break;

      case STRSW_str_10:

	D->type = typ_str;
	break;
      default:
	exit (fprintf (stderr, "unknown type in "
		       "declaration at row %d : %s\n", i + 1, C));

      }
  }

  j++;
  D->format = get_cell (G, i, j);
  j++;
  C = get_cell (G, i, j);
  if (!C)
    exit (fprintf (stderr, "empty editable field in "
		   "declaration at row %d\n", i + 1));
  {
    int _cmp14;
    int _swvar14;
    char *_strvar14;

    enum case_values
    {
      __UNKNOWN_14,

      STRSW_y_14,

      STRSW_Y_14,

      STRSW_n_14,

      STRSW_N_14,

    };

    _strvar14 = C;

    _swvar14 = __UNKNOWN_14;

    _cmp14 = strcmp (_strvar14, "Y");
    if (_cmp14 < 0)
      {
	_cmp14 = strcmp (_strvar14, "N");
	if (_cmp14 == 0)
	  {
	    _swvar14 = STRSW_N_14;
	  }
      }
    else if (_cmp14 > 0)
      {
	_cmp14 = strcmp (_strvar14, "n");
	if (_cmp14 < 0)
	  {
	  }
	else if (_cmp14 > 0)
	  {
	    _cmp14 = strcmp (_strvar14, "y");
	    if (_cmp14 == 0)
	      {
		_swvar14 = STRSW_y_14;
	      }
	  }
	else
	  {
	    _swvar14 = STRSW_n_14;
	  }
      }
    else
      {
	_swvar14 = STRSW_Y_14;
      }

    switch (_swvar14)
      {
/* I remove this because the default case doesnt work otherwise..
		case __UNKNOWN_14:
			break;
*/

      case STRSW_y_14:

      case STRSW_Y_14:

	D->editable = 1;
	break;

      case STRSW_n_14:

      case STRSW_N_14:

	D->editable = 0;
	break;
      default:
	exit (fprintf (stderr, "unknown editable field in "
		       "declaration at row %d\n", i + 1));

      }
  }

  j++;
  C = get_cell (G, i, j);
  D->show_if_zero = 0;
  if (C && D->type == typ_int)
    {
      {
	int _cmp20;
	int _swvar20;
	char *_strvar20;

	enum case_values
	{
	  __UNKNOWN_20,

	  STRSW_y_20,

	  STRSW_Y_20,

	};

	_strvar20 = C;

	_swvar20 = __UNKNOWN_20;

	_cmp20 = strcmp (_strvar20, "Y");
	if (_cmp20 > 0)
	  {
	    _cmp20 = strcmp (_strvar20, "y");
	    if (_cmp20 == 0)
	      {
		_swvar20 = STRSW_y_20;
	      }
	  }
	else if (_cmp20 == 0)
	  {
	    _swvar20 = STRSW_Y_20;
	  }

	switch (_swvar20)
	  {
/* I remove this because the default case doesnt work otherwise..
		case __UNKNOWN_20:
			break;
*/

	  case STRSW_y_20:

	  case STRSW_Y_20:

	    D->show_if_zero = 1;
	    break;

	  }
      }

    }
  declist_append (L, D);
}

static char *
get_cell (grid_t * G, int r, int c)
{
  char *data;
  if (r < 0 || r >= G->nrows)
    return 0;
  if (c < 0 || c >= G->ncols)
    return 0;
  if (!G->cells[r][c].data)
    return 0;
  data = strdup (G->cells[r][c].data);
  remove_spaces (data);
  return data;
}
