/*

Copyright (c) 2009 Şükrü Çınar  (sukru cinar)

This program is distributed under GNU GPL v2. See file COPYING
for licensing details.

*/

#include <stdlib.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "parsedoc.h"
#include "style.h"
#include "decl.h"
#include "miscfn.h"
#include "repstr.h"

#include "field.h"
typedef struct
{
  int r1, c1, r2, c2;
} rect_t;
static void print_field_enum (cbuffer_t * buffer, field_t * F);
static void print_destruct_code (cbuffer_t * buffer, field_t * F);
static void print_init_code (cbuffer_t * buffer, field_t * F);
static void print_call_to_editing_callback (cbuffer_t * buffer, field_t * F);
static void print_declaration (cbuffer_t * buffer, field_t * F);
static void print_callback_declaration (cbuffer_t * buffer, field_t * F);
static void remove_spaces_from_data_cells (grid_t * G);
static void compute_field (grid_t * G, decl_t * D, storage_t * S);
static int get_decl_rect (grid_t * G, decl_t * D, rect_t * R);
static void cellreflist_appendC (cellreflist_t * R, int a, int b);
static void print_fields
  (fieldlist_t * fields, void (*func) (cbuffer_t *, field_t *),
   char *vname, strpairlist_t * vars);
static void print_raw_declaration (cbuffer_t * buffer, field_t * F);
static void print_raw_destructor (cbuffer_t * buffer, field_t * F);

void *__unused_field_c[] = {

  0,
};

void cellreflist_init (cellreflist_t * L)
{
  L->head.prev = L->head.next = &L->head;
}

cellref_t *cellreflist_destroy (cellreflist_t * L)
{
  typeof (L->head.next) N;
  cellref_t *R;

  N = L->head.next;
  if (N == &L->head)
    return 0;
  R = N->data;

  cellreflist_remove (L, N);
  return R;
}

cellreflist_t *cellreflist_new ()
{
  cellreflist_t *L;
  L = (typeof (L)) malloc (sizeof (*L));
  L->head.prev = L->head.next = &L->head;
  return L;
}

void cellreflist_append (cellreflist_t * L, cellref_t * data)
{
  struct cellreflist_t_NODE *n;
  n = (typeof (n)) malloc (sizeof (n[0]));
  n->data = data;
  n->prev = L->head.prev;
  n->next = &L->head;
  L->head.prev->next = n;
  L->head.prev = n;
}

void cellreflist_prepend (cellreflist_t * L, cellref_t * data)
{
  struct cellreflist_t_NODE *n;
  n = (typeof (n)) malloc (sizeof (n[0]));
  n->data = data;
  n->next = L->head.next;
  n->prev = &L->head;
  L->head.next->prev = n;
  L->head.next = n;
}

void cellreflist_remove (cellreflist_t * L, struct cellreflist_t_NODE *n)
{
  n->prev->next = n->next;
  n->next->prev = n->prev;
  free (n);
}

int cellreflist_isempty (cellreflist_t * L)
{
  return L->head.next == &L->head;
}

void fieldlist_init (fieldlist_t * L)
{
  L->head.prev = L->head.next = &L->head;
}

field_t *fieldlist_destroy (fieldlist_t * L)
{
  typeof (L->head.next) N;
  field_t *R;

  N = L->head.next;
  if (N == &L->head)
    return 0;
  R = N->data;

  fieldlist_remove (L, N);
  return R;
}

fieldlist_t *fieldlist_new ()
{
  fieldlist_t *L;
  L = (typeof (L)) malloc (sizeof (*L));
  L->head.prev = L->head.next = &L->head;
  return L;
}

void fieldlist_append (fieldlist_t * L, field_t * data)
{
  struct fieldlist_t_NODE *n;
  n = (typeof (n)) malloc (sizeof (n[0]));
  n->data = data;
  n->prev = L->head.prev;
  n->next = &L->head;
  L->head.prev->next = n;
  L->head.prev = n;
}

void fieldlist_prepend (fieldlist_t * L, field_t * data)
{
  struct fieldlist_t_NODE *n;
  n = (typeof (n)) malloc (sizeof (n[0]));
  n->data = data;
  n->next = L->head.next;
  n->prev = &L->head;
  L->head.next->prev = n;
  L->head.next = n;
}

void fieldlist_remove (fieldlist_t * L, struct fieldlist_t_NODE *n)
{
  n->prev->next = n->next;
  n->next->prev = n->prev;
  free (n);
}

int fieldlist_isempty (fieldlist_t * L)
{
  return L->head.next == &L->head;
}

static struct
{
  char *name;
  void (*func) (cbuffer_t *, field_t *);
} field_printers[] =
{
  {
  "$FIELD_ENUMS", print_field_enum},
  {
  "$STORAGE_DECLARATIONS", print_declaration},
  {
  "$CELL_CHANGE_CALLBACK_DECLARATIONS", print_callback_declaration},
  {
  "$CELL_CHANGE_CALLBACK_CASES", print_call_to_editing_callback},
  {
  "$STORAGE_INIT_CODE", print_init_code},
  {
  "$STORAGE_DESTRUCTOR_CODE", print_destruct_code},
  {
  "$RAW_STORAGE_DECLARATIONS", print_raw_declaration},
  {
  "$RAW_DESTRUCTOR", print_raw_destructor},
  {
  0, 0}
};

storage_t *compute_fields
  (grid_t * layout, declist_t * decs, strpairlist_t * vars)
{
  int i;
  storage_t *S;

  S = (typeof (S)) calloc (1, sizeof (S[0]));

  S->fields = fieldlist_new ();
  S->nrows = layout->nrows;
  S->ncols = layout->ncols;

  S->field_map =
    (typeof (S->field_map)) calloc (S->nrows, sizeof (S->field_map[0]));

  for (i = 0; i < S->nrows; i++)

    S->field_map[i] =
      (typeof (S->field_map[i])) calloc (S->ncols,
					 sizeof (S->field_map[i][0]));

  remove_spaces_from_data_cells (layout);
  {
    typeof ((*decs).head.next) N15, T15;
    for (T15 = (*decs).head.next; T15 != &((*decs).head); T15 = N15)
      {
	typeof (T15->data) D;
	D = T15->data;
	N15 = T15->next;

	compute_field (layout, D, S);

      }
  }

  for (i = 0; field_printers[i].name; i++)
    print_fields (S->fields, field_printers[i].func,
		  field_printers[i].name, vars);

  return S;
}

static void
print_field_enum (cbuffer_t * buffer, field_t * F)
{
  cbprintf (buffer, " field_%s, ", F->name);
}

static void
print_destruct_code (cbuffer_t * buffer, field_t * F)
{
  if (F->type != typ_str)
    return;
  switch (F->kind)
    {
    case field_single:
      cbprintf (buffer, "freeifnnull(S.%s.value);\n", F->name);
      break;
    case field_2darray:
      cbprintf (buffer, "\tfor(i=0;i<%d;i++)\n", F->dim1);
      cbprintf (buffer, "\t\tfor(j=0;i<%d;i++)\n", F->dim2);
      cbprintf (buffer, "\t\t\tfreeifnnull(S.%s[i][j].value);\n", F->name);
      break;
    case field_sarray:
    case field_array:
      cbprintf (buffer, "\tfor(i=0;i<%d;i++)\n", F->dim1);
      cbprintf (buffer, "\t\tfreeifnnull(S.%s[i].value);\n", F->name);
      break;
    }
}

static void
print_init_code (cbuffer_t * buffer, field_t * F)
{
  int i;
  switch (F->kind)
    {
    case field_single:
      cbprintf (buffer, "\tS.%s.init(this, %d,%d);\n", F->name, F->r1, F->c1);
      break;
    case field_2darray:
      cbprintf (buffer, "\tfor(i=0;i<%d;i++)\n", F->dim1);
      cbprintf (buffer, "\t\tfor(j=0;i<%d;i++)\n", F->dim2);
      cbprintf (buffer, "\t\t\tS.%s[i][j].init(this, i+%d,j+%d);\n",
		F->name, F->r1, F->c1);
      break;
    case field_array:
      cbprintf (buffer, "\tfor(i=0;i<%d;i++)\n", F->dim1);
      if (F->r1 == F->r2)
	cbprintf (buffer, "\t\tS.%s[i].init(this, %d,i+%d);\n",
		  F->name, F->r1, F->c1);
      else
	cbprintf (buffer, "\t\tS.%s[i].init(this, i+%d,%d);\n",
		  F->name, F->r1, F->c1);
      break;
    case field_sarray:
      i = 0;
      {
	typeof ((*F->cells).head.next) N20, T20;
	for (T20 = (*F->cells).head.next;
	     T20 != &((*F->cells).head); T20 = N20)
	  {
	    typeof (T20->data) C;
	    C = T20->data;
	    N20 = T20->next;

	    cbprintf (buffer, "\tS.%s[%d].init(this, %d,%d);\n",
		      F->name, i, C->r, C->c);
	    i++;

	  }
      }

    }
}

static void print_call_to_editing_callback (cbuffer_t * buffer, field_t * F)
{
  if (!F->editable)
    {
      cbprintf (buffer,
		"\tcase field_%s: /* non-editable */ break;\n", F->name);
      return;
    }
  cbprintf (buffer, "\tcase field_%s: onccr_%s(", F->name, F->name);
  switch (F->type)
    {
    case typ_int:
      cbprintf (buffer, "intv");
      break;
    case typ_str:
      cbprintf (buffer, "strv");
      break;
    }
  if (F->kind != field_single)
    cbprintf (buffer, ",idx1");
  if (F->kind == field_2darray)
    cbprintf (buffer, ",idx2");
  cbprintf (buffer, "); break;\n");
}

static void
print_declaration (cbuffer_t * buffer, field_t * F)
{
  cbprintf (buffer, "\t");
  switch (F->type)
    {
    case typ_int:
      cbprintf (buffer, "boxed_int ");
      break;
    case typ_str:
      cbprintf (buffer, "boxed_str ");
      break;
    }

  cbprintf (buffer, "%s", F->name);

  switch (F->kind)
    {
    case field_single:
      break;
    case field_2darray:
      cbprintf (buffer, "[%d][%d]", F->dim1, F->dim2);
      break;
    default:
      cbprintf (buffer, "[%d]", F->dim1);
      break;
    }
  cbprintf (buffer, ";\n");
}

static void print_callback_declaration (cbuffer_t * buffer, field_t * F)
{
  cbprintf (buffer, "\t void onccr_%s(", F->name);
  switch (F->type)
    {
    case typ_int:
      cbprintf (buffer, "int value");
      break;
    case typ_str:
      cbprintf (buffer, "char *value");
      break;
    }
  switch (F->kind)
    {
    case field_sarray:
    case field_array:
      cbprintf (buffer, ",int idx");
      break;
    case field_2darray:
      cbprintf (buffer, ",int idx1,int idx2");
      break;
    default:
      break;
    }
  cbprintf (buffer, ");\n");
}

static void
remove_spaces_from_data_cells (grid_t * G)
{
  int r, c;
  for (r = 0; r < G->nrows; r++)
    for (c = 0; c < G->ncols; c++)
      if (G->cells[r][c].kind == ck_data)
	{
	  if (G->cells[r][c].data)
	    {
	      remove_spaces (G->cells[r][c].data);
	      if (!G->cells[r][c].data[0])
		G->cells[r][c].data = 0;
	    }
	}
}

static void
compute_field (grid_t * G, decl_t * D, storage_t * S)
{
  int i, j;
  rect_t R;
  field_t *F;
  int find;

  find = get_decl_rect (G, D, &R);
  if (find < 0)
    return;

  F = (typeof (F)) calloc (1, sizeof (F[0]));

  F->cells = cellreflist_new ();
  fieldlist_append (S->fields, F);

  if (find)
    {
      if (R.r2 != R.r1 && R.c2 != R.c1)
	{
	  F->kind = field_2darray;
	  F->r1 = R.r1;
	  F->r2 = R.r2;
	  F->c1 = R.c1;
	  F->c2 = R.c2;
	  F->dim1 = F->r2 - F->r1 + 1;
	  F->dim2 = F->c2 - F->c1 + 1;
	  for (i = R.r1; i <= R.r2; i++)
	    for (j = R.c1; j <= R.c2; j++)
	      {
		cellreflist_appendC (F->cells, i, j);

		S->field_map[i][j].field = F;
		S->field_map[i][j].idx1 = i - R.r1;
		S->field_map[i][j].idx2 = j - R.c1;
	      }
	}
      else if (R.r2 == R.r1 && R.c2 == R.c2)
	{
	  cellreflist_appendC (F->cells, R.r1, R.c1);

	  F->kind = field_single;
	  F->r1 = R.r1;
	  F->c1 = R.c1;
	  S->field_map[R.r1][R.c1].field = F;
	}
      else
	{
	  F->kind = field_array;
	  F->r1 = R.r1;
	  F->r2 = R.r2;
	  F->c1 = R.c1;
	  F->c2 = R.c2;
	  if (R.r1 == R.r2)
	    F->dim1 = R.c2 - R.c1 + 1;
	  else
	    F->dim1 = R.r2 - R.r1 + 1;
	  for (i = R.r1; i <= R.r2; i++)
	    for (j = R.c1; j <= R.c2; j++)
	      {
		cellreflist_appendC (F->cells, i, j);

		S->field_map[i][j].field = F;
		S->field_map[i][j].idx1 = i - R.r1 + j - R.c1;
	      }
	}
    }
  else
    {
      F->kind = field_sarray;
      F->dim1 = 0;
      for (i = 0; i < G->nrows; i++)
	for (j = 0; j < G->ncols; j++)
	  {
	    if (G->cells[i][j].kind != ck_data)
	      continue;
	    if (!G->cells[i][j].data)
	      continue;
	    if (strcmp (G->cells[i][j].data, D->name))
	      continue;

	    cellreflist_appendC (F->cells, i, j);

	    S->field_map[i][j].field = F;
	    S->field_map[i][j].idx1 = F->dim1;
	    F->dim1++;
	  }
    }
  F->name = strdup (D->name);
  if (D->format)
    {
      F->format = strdup (D->format);
      remove_spaces (F->format);
      if (!F->format[0])
	F->format = 0;
    }
  else
    F->format = 0;

  F->type = D->type;
  F->editable = D->editable;
  F->show_if_zero = D->show_if_zero;
}

static int
get_decl_rect (grid_t * G, decl_t * D, rect_t * R)
{
  int max_r, min_r, max_c, min_c;
  int r, c;
  int count;

  max_r = -1;
  min_r = G->nrows;
  max_c = -1;
  min_c = G->ncols;
  count = 0;

  for (r = 0; r < G->nrows; r++)
    for (c = 0; c < G->ncols; c++)
      {
	if (G->cells[r][c].kind != ck_data)
	  continue;
	if (!G->cells[r][c].data)
	  continue;
	if (strcmp (G->cells[r][c].data, D->name))
	  continue;
	count++;
	if (r > max_r)
	  max_r = r;
	if (r < min_r)
	  min_r = r;
	if (c > max_c)
	  max_c = c;
	if (c < min_c)
	  min_c = c;
      }

  if (max_r == -1)
    return -1;
  if (count != ((max_r - min_r + 1) * (max_c - min_c + 1)))
    return 0;
  R->r1 = min_r;
  R->r2 = max_r;
  R->c1 = min_c;
  R->c2 = max_c;
  return 1;
}

static void
cellreflist_appendC (cellreflist_t * R, int a, int b)
{
  cellref_t *C;

  C = (typeof (C)) calloc (1, sizeof (C[0]));

  C->r = a;
  C->c = b;
  cellreflist_append (R, C);
}

static void print_fields
  (fieldlist_t * fields, void (*func) (cbuffer_t *, field_t *),
   char *vname, strpairlist_t * vars)
{
  int dummy;
  cbuffer_t *buffer;
  buffer = cbuffer_new ();
  {
    typeof ((*fields).head.next) N33, T33;
    for (T33 = (*fields).head.next; T33 != &((*fields).head); T33 = N33)
      {
	typeof (T33->data) F;
	F = T33->data;
	N33 = T33->next;

	(*func) (buffer, F);

      }
  }

  cbuffer_append (buffer, 0);
  strpairlist_append (vars, strpair_new (strdup (vname),
					 cbuffer_collectda (buffer, &dummy)));
}

static void print_raw_declaration (cbuffer_t * buffer, field_t * F)
{
  cbprintf (buffer, "\t");
  switch (F->type)
    {
    case typ_int:
      cbprintf (buffer, "int ");
      break;
    case typ_str:
      cbprintf (buffer, "char* ");
      break;
    }
  cbprintf (buffer, "%s", F->name);
  if (F->kind != field_single)
    cbprintf (buffer, "[%d]", F->dim1);
  if (F->kind == field_2darray)
    cbprintf (buffer, "[%d]", F->dim2);
  cbprintf (buffer, ";\n");
}

static void print_raw_destructor (cbuffer_t * buffer, field_t * F)
{
  if (F->type != typ_str)
    return;
  switch (F->kind)
    {
    case field_single:
      cbprintf (buffer, "freeifnnull(obj->%s);\n", F->name);
      break;
    case field_2darray:
      cbprintf (buffer, "\t{int i,j; for(i=0;i<%d;i++)\n", F->dim1);
      cbprintf (buffer, "\t\tfor(j=0;i<%d;i++)\n", F->dim2);
      cbprintf (buffer, "\t\t\tfreeifnnull(obj->%s[i][j]);}\n", F->name);
      break;
    case field_sarray:
    case field_array:
      cbprintf (buffer, "\t{int i; for(i=0;i<%d;i++)\n", F->dim1);
      cbprintf (buffer, "\t\tfreeifnnull(obj->%s[i]);}\n", F->name);
      break;
    }
}
