/*

Copyright (c) 2009 Şükrü Çınar  (sukru cinar)

This program is distributed under GNU GPL v2. See file COPYING
for licensing details.

*/

#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "repstr.h"
#include "miscfn.h"
#include "parsedoc.h"
#include "style.h"
#include "decl.h"
#include "field.h"

#include "gencells.h"

static void print_cell_def (xcell_t * C, cbuffer_t * buffer, storage_t * S);
static void print_format (cbuffer_t * buffer, char *format);
static void print_c_literal (cbuffer_t * buffer, char *D);
static void print_cell_data (cbuffer_t * buffer, xcell_t * C);
static char *align_to_valign (int a);
static char *align_to_halign (int a);
static char *base_type_str (enum celltype typ);

void *__unused_gencells_c[] = {

  0,
};

void
gen_cells (grid_t * G, storage_t * S, strpairlist_t * vars)
{
  cbuffer_t *buffer;
  int dummy;
  int i, j;

  buffer = cbuffer_new ();
  for (i = 0; i < G->nrows; i++)
    {
      cbuffer_append (buffer, '{');
      for (j = 0; j < G->ncols; j++)
	print_cell_def (&G->cells[i][j], buffer, S);
      cbuffer_putstr (buffer, "},\n");
    }
  cbuffer_append (buffer, 0);
  strpairlist_append (vars,
		      strpair_new (strdup ("$CELL_DEFS"),
				   cbuffer_collectda (buffer, &dummy)));

  buffer = cbuffer_new ();
  for (i = 0; i < G->nrows; i++)
    for (j = 0; j < G->ncols; j++)
      if (G->cells[i][j].kind == ck_header)
	print_cell_data (buffer, &G->cells[i][j]);

  cbuffer_append (buffer, 0);
  strpairlist_append (vars,
		      strpair_new (strdup ("$CELL_INIT_CODE"),
				   cbuffer_collectda (buffer, &dummy)));
}

static void print_cell_def (xcell_t * C, cbuffer_t * buffer, storage_t * S)
{
  field_t *F;
  switch (C->kind)
    {
    case ck_merged:
      cbprintf (buffer, "{ck_merged, 0,0, 0,0, %d, %d, 0},\n",
		C->merged_to->row, C->merged_to->col);
      break;
    case ck_empty:
      cbprintf (buffer, "{ck_empty,0, 0, %d, %d, 0,0,0},\n",
		C->rowspan, C->colspan);
      break;
    case ck_header:
      cbprintf (buffer, "{ck_header, %s, %s, %d, %d, 0,0,0},\n",
		align_to_halign (C->halign),
		align_to_valign (C->valign), C->rowspan, C->colspan);
      break;
    case ck_data:
      F = S->field_map[C->row][C->col].field;
      cbprintf (buffer, "{ck_data, %s, %s, %d, %d, 0,0,\n",
		align_to_halign (C->halign),
		align_to_valign (C->valign), C->rowspan, C->colspan);
      cbprintf (buffer, " %d, %s, field_%s, %d, %d, \n",
		F->editable,
		base_type_str (F->type),
		F->name,
		S->field_map[C->row][C->col].idx1,
		S->field_map[C->row][C->col].idx2);
      print_format (buffer, F->format);
      cbprintf (buffer, ",%d,",
		S->field_map[C->row][C->col].field->show_if_zero);
      cbprintf (buffer, "},\n");
      break;
    }
}

static void
print_format (cbuffer_t * buffer, char *format)
{
  if (!format)
    goto empty;
  format = strdup (format);
  if (!format[0])
    goto empty;
  print_c_literal (buffer, format);
  return;

empty:
  cbprintf (buffer, "0");
}

static void
print_c_literal (cbuffer_t * buffer, char *D)
{
  int i;

  cbuffer_append (buffer, '"');
  for (i = 0; D[i]; i++)
    {
      char xb[20];
      if (D[i] >= '#' && D[i] < '\\')
	cbuffer_append (buffer, D[i]);
      else if (D[i] >= ']' && D[i] <= '~')
	cbuffer_append (buffer, D[i]);
      else
	{
	  sprintf (xb, "\\%o", D[i]);
	  cbuffer_putstr (buffer, xb);
	}
    }
  cbuffer_append (buffer, '"');
}

static void
print_cell_data (cbuffer_t * buffer, xcell_t * C)
{
  cbprintf (buffer, "\tcells[%d][%d].data=strdup(", C->row, C->col);
  print_c_literal (buffer, C->data);
  cbprintf (buffer, ");\n");
}

static char *
align_to_valign (int a)
{
  switch (a)
    {
    case al_begin:
      return "valign_top";
    case al_end:
      return "valign_bottom";
    case al_center:
      return "valign_center";
    }
  return "valign_bottom";
}

static char *
align_to_halign (int a)
{
  switch (a)
    {
    case al_begin:
      return "halign_left";
    case al_end:
      return "halign_right";
    case al_center:
      return "halign_center";
    }
  return "halign_left";
}

static char *
base_type_str (enum celltype typ)
{
  switch (typ)
    {
    case typ_int:
      return strdup ("cell_int");
    case typ_str:
      return strdup ("cell_str");
    }
  return strdup ("unknown base type");
}
