/*

Copyright (c) 2009 Şükrü Çınar  (sukru cinar)

This program is distributed under GNU GPL v2. See file COPYING
for licensing details.

*/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "parsedoc.h"
#include "repstr.h"
#include "style.h"

#include "genlines.h"
typedef struct
{
  int a;
  int b1;
  int b2;
} sline_t;
struct cbuffer_t_node
{
  struct cbuffer_t_node *next;
  int len;
  char data[1000];
};
typedef struct
{
  struct cbuffer_t_node *first;
  struct cbuffer_t_node *last;
} cbuffer_t;
struct sbuffer_t_node
{
  struct sbuffer_t_node *next;
  int len;
  sline_t data[200];
};
typedef struct
{
  struct sbuffer_t_node *first;
  struct sbuffer_t_node *last;
} sbuffer_t;
static void cbuffer_init (cbuffer_t * B);
static cbuffer_t *cbuffer_new ();
static void cbuffer_append (cbuffer_t * B, char D);
static char *cbuffer_collect (cbuffer_t * B, int *size);
static char *cbuffer_collectd (cbuffer_t * B, int *size);
static char *cbuffer_collectda (cbuffer_t * B, int *size);
static void cbuffer_destroy (cbuffer_t * B);
static void cbuffer_destroya (cbuffer_t * B);
static void sbuffer_init (sbuffer_t * B);
static sbuffer_t *sbuffer_new ();
static void sbuffer_append (sbuffer_t * B, sline_t D);
static sline_t *sbuffer_collect (sbuffer_t * B, int *size);
static sline_t *sbuffer_collectd (sbuffer_t * B, int *size);
static sline_t *sbuffer_collectda (sbuffer_t * B, int *size);
static void sbuffer_destroy (sbuffer_t * B);
static void sbuffer_destroya (sbuffer_t * B);
static void do_lines
  (sbuffer_t * buffer, char *name_n, char *name_def, strpairlist_t * vars);
static void put_lines (cbuffer_t * buffer, sline_t * lines, int nlines);
static sline_t *combine_lines (sline_t * lines, int *n);

static void
quicksort_13 (sline_t * A, int left, int right)
{
  if (right <= left)
    return;
  /* pivot position */
  int ppos = left;
  int i;
  for (i = left; i < right; i++)
    {
      /* this will have the effect of no-sort
         if a comparision is not given */
      int result = -1;
      sline_t *X1 = A + i;
      sline_t *X2 = A + right;

      result = X1->a - X2->a;
      result = (result == 0) ? (X1->b1 - X2->b1) : result;
      result = (result == 0) ? (X1->b2 - X2->b2) : result;

      if (result <= 0)
	{
	  sline_t V1 = A[i];
	  A[i] = A[ppos];
	  A[ppos] = V1;
	  ppos++;
	}
    }
  {
    sline_t V2 = A[ppos];
    A[ppos] = A[right];
    A[right] = V2;
  }
  quicksort_13 (A, left, ppos - 1);
  quicksort_13 (A, ppos + 1, right);
}

void *__unused_genlines_c[] = {

  (void *) cbuffer_collectd,

  (void *) sbuffer_collectd,

  0,
};

static void cbuffer_init (cbuffer_t * B)
{
  B->first = malloc (sizeof (*B->first));
  B->first->next = 0;
  B->first->len = 0;
  B->last = B->first;
}

static cbuffer_t *cbuffer_new ()
{
  cbuffer_t *B;
  B = malloc (sizeof (*B));

  cbuffer_init (B);
  return B;
}

static void cbuffer_append (cbuffer_t * B, char D)
{
  struct cbuffer_t_node *L;
  L = B->last;
  if (L->len == 1000)
    {
      L = malloc (sizeof (*L));
      L->len = 0;
      L->next = 0;
      B->last->next = L;
      B->last = L;
    }
  L->data[L->len] = D;
  L->len++;
}

static char *cbuffer_collect (cbuffer_t * B, int *size)
{
  struct cbuffer_t_node *L;
  int C;
  char *R;

  {
    C = 0;
    for (L = B->first; L; L = L->next)
      C += L->len;
  }

  if (size)
    *size = C;
  if (C)
    {
      R = malloc (C * sizeof (*R));

      {
	C = 0;
	for (L = B->first; L; L = L->next)
	  {
	    int i;
	    for (i = 0; i < L->len; i++)
	      R[C + i] = L->data[i];
	    C += L->len;
	  }
      }

    }
  else
    R = 0;
  return R;
}

static char *cbuffer_collectd (cbuffer_t * B, int *size)
{
  char *R;
  R = cbuffer_collect (B, size);

  cbuffer_destroy (B);
  return R;
}

static char *cbuffer_collectda (cbuffer_t * B, int *size)
{
  char *R;
  R = cbuffer_collect (B, size);

  cbuffer_destroya (B);
  return R;
}

static void cbuffer_destroy (cbuffer_t * B)
{
  struct cbuffer_t_node *L, *N;

  for (L = B->first; L; L = N)
    {
      N = L->next;
      free (L);
    }
  B->first = B->last = 0;
}

static void cbuffer_destroya (cbuffer_t * B)
{

  cbuffer_destroy (B);
  free (B);
}

static void sbuffer_init (sbuffer_t * B)
{
  B->first = malloc (sizeof (*B->first));
  B->first->next = 0;
  B->first->len = 0;
  B->last = B->first;
}

static sbuffer_t *sbuffer_new ()
{
  sbuffer_t *B;
  B = malloc (sizeof (*B));

  sbuffer_init (B);
  return B;
}

static void sbuffer_append (sbuffer_t * B, sline_t D)
{
  struct sbuffer_t_node *L;
  L = B->last;
  if (L->len == 200)
    {
      L = malloc (sizeof (*L));
      L->len = 0;
      L->next = 0;
      B->last->next = L;
      B->last = L;
    }
  L->data[L->len] = D;
  L->len++;
}

static sline_t *sbuffer_collect (sbuffer_t * B, int *size)
{
  struct sbuffer_t_node *L;
  int C;
  sline_t *R;

  {
    C = 0;
    for (L = B->first; L; L = L->next)
      C += L->len;
  }

  if (size)
    *size = C;
  if (C)
    {
      R = malloc (C * sizeof (*R));

      {
	C = 0;
	for (L = B->first; L; L = L->next)
	  {
	    int i;
	    for (i = 0; i < L->len; i++)
	      R[C + i] = L->data[i];
	    C += L->len;
	  }
      }

    }
  else
    R = 0;
  return R;
}

static sline_t *sbuffer_collectd (sbuffer_t * B, int *size)
{
  sline_t *R;
  R = sbuffer_collect (B, size);

  sbuffer_destroy (B);
  return R;
}

static sline_t *sbuffer_collectda (sbuffer_t * B, int *size)
{
  sline_t *R;
  R = sbuffer_collect (B, size);

  sbuffer_destroya (B);
  return R;
}

static void sbuffer_destroy (sbuffer_t * B)
{
  struct sbuffer_t_node *L, *N;

  for (L = B->first; L; L = N)
    {
      N = L->next;
      free (L);
    }
  B->first = B->last = 0;
}

static void sbuffer_destroya (sbuffer_t * B)
{

  sbuffer_destroy (B);
  free (B);
}

void
gen_lines (grid_t * G, strpairlist_t * vars)
{
  int i, j;
  sbuffer_t *buffer;

  buffer = sbuffer_new ();

  for (i = 0; i < G->nrows; i++)
    for (j = 0; j < G->ncols; j++)
      {
	xcell_t *X;
	X = &G->cells[i][j];
	if (X->kind == ck_merged)
	  continue;
	if (!X->colspan || !X->rowspan)
	  continue;
	if (X->border & b_bottom)
	  {
	    sline_t L;
	    L.a = i + X->rowspan;
	    L.b1 = j;
	    L.b2 = j + X->colspan;
	    sbuffer_append (buffer, L);
	  }
	if (X->border & b_top)
	  {
	    sline_t L;
	    L.a = i;
	    L.b1 = j;
	    L.b2 = j + X->colspan;
	    sbuffer_append (buffer, L);
	  }
      }

  do_lines (buffer, "$NHORZLINES", "$HORZ_LINE_DEFS", vars);

  buffer = sbuffer_new ();

  for (j = 0; j < G->ncols; j++)
    for (i = 0; i < G->nrows; i++)
      {
	xcell_t *X;
	X = &G->cells[i][j];
	if (X->kind == ck_merged)
	  continue;
	if (!X->colspan || !X->rowspan)
	  continue;
	if (X->border & b_left)
	  {
	    sline_t L;
	    L.a = j;
	    L.b1 = i;
	    L.b2 = i + X->rowspan;
	    sbuffer_append (buffer, L);
	  }
	if (X->border & b_right)
	  {
	    sline_t L;
	    L.a = j + 1;
	    L.b1 = i;
	    L.b2 = i + X->rowspan;
	    sbuffer_append (buffer, L);
	  }
      }

  do_lines (buffer, "$NVERTLINES", "$VERT_LINE_DEFS", vars);
}

static void do_lines
  (sbuffer_t * buffer, char *name_n, char *name_def, strpairlist_t * vars)
{
  cbuffer_t *obuffer;
  int nlines;
  sline_t *lines;
  char tbuf[20];
  int dummy;

  lines = sbuffer_collectda (buffer, &nlines);
  lines = combine_lines (lines, &nlines);
  sprintf (tbuf, "%d", nlines);
  strpairlist_append (vars, strpair_new (strdup (name_n), strdup (tbuf)));

  obuffer = cbuffer_new ();
  put_lines (obuffer, lines, nlines);
  free (lines);
  strpairlist_append (vars,
		      strpair_new (strdup (name_def),
				   cbuffer_collectda (obuffer, &dummy)));
}

static void
put_lines (cbuffer_t * buffer, sline_t * lines, int nlines)
{
  int i;
  char xb[1000];
  for (i = 0; i < nlines; i++)
    {
      int j;
      sprintf (xb, "{ %d, %d, %d },\n", lines[i].a, lines[i].b1, lines[i].b2);
      for (j = 0; xb[j]; j++)
	cbuffer_append (buffer, xb[j]);
    }
  cbuffer_append (buffer, 0);
}

static sline_t *
combine_lines (sline_t * lines, int *n)
{
  int N;

  N = *n;

  quicksort_13 (lines, 0, N - 1);

  do
    {
      *n = N;

      {
	int i = 0, j = 0;
	int result;
	while (i < (N) && j < (N))
	  {

	    result = 1;
	    if (lines[i].a == lines[j].a)
	      {
		if (lines[i].b1 >= lines[j].b1 && lines[i].b1 < lines[j].b2)
		  {
		    if (lines[j].b2 > lines[i].b2)
		      lines[i].b2 = lines[j].b2;
		    lines[i].b1 = lines[j].b1;
		    result = 0;
		  }
		else if (lines[i].b2 >= lines[j].b1
			 && lines[i].b2 < lines[j].b2)
		  {
		    if (lines[j].b1 < lines[i].b1)
		      lines[i].b1 = lines[j].b1;
		    lines[i].b2 = lines[j].b2;
		    result = 0;
		  }
	      }

	    if (result)
	      {
		i++;
		lines[i] = lines[j];
	      }
	    j++;
	  }
	N = i + 1;
      }

    }
  while (N != *n);

  return lines;
}
