/*

Copyright (c) 2009 Şükrü Çınar  (sukru cinar)

This program is distributed under GNU GPL v2. See file COPYING
for licensing details.

*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "miscfn.h"
#include "repstr.h"
#include "parsedoc.h"
#include "style.h"
#include "decl.h"
#include "field.h"

#include "serialize.h"

static void output_code
  (strpairlist_t * vars, fieldlist_t * fields, char *vname,
   void (*func) (field_t *, cbuffer_t *));
static void serialize_field (field_t * F, cbuffer_t * B);
static void deserialize_field (field_t * F, cbuffer_t * B);
static void call_method_on_field (field_t * F, cbuffer_t * B, char *M);
static void serialize_raw (field_t * F, cbuffer_t * B);
static void deserialize_raw (field_t * F, cbuffer_t * B);

void *__unused_serialize_c[] = {

  0,
};

void output_serialization_code (storage_t * S, strpairlist_t * vars)
{
  output_code (vars, S->fields, "$SERIALIZATION", serialize_field);
  output_code (vars, S->fields, "$DESERIALIZATION", deserialize_field);
  output_code (vars, S->fields, "$RAW_PARSING_CODE", deserialize_raw);
  output_code (vars, S->fields, "$RAW_UNPARSING_CODE", serialize_raw);
}

static void output_code
  (strpairlist_t * vars, fieldlist_t * fields, char *vname,
   void (*func) (field_t *, cbuffer_t *))
{
  cbuffer_t *buffer;
  int dummy;
  buffer = cbuffer_new ();
  {
    typeof ((*fields).head.next) N6, T6;
    for (T6 = (*fields).head.next; T6 != &((*fields).head); T6 = N6)
      {
	typeof (T6->data) F;
	F = T6->data;
	N6 = T6->next;

	(*func) (F, buffer);

      }
  }

  cbuffer_append (buffer, 0);
  strpairlist_append (vars, strpair_new (strdup (vname),
					 cbuffer_collectda (buffer, &dummy)));
}

static void
serialize_field (field_t * F, cbuffer_t * B)
{
  call_method_on_field (F, B, "serialize");
}

static void
deserialize_field (field_t * F, cbuffer_t * B)
{
  call_method_on_field (F, B, "deserialize");
}

static void
call_method_on_field (field_t * F, cbuffer_t * B, char *M)
{
  switch (F->kind)
    {
    case field_single:
      cbprintf (B, "\tS.%s.%s(B,L);\n", F->name, M);
      break;
    case field_array:
    case field_sarray:
      cbprintf (B, "\t{int i;for(i=0;i<%d;i++)\n", F->dim1);
      cbprintf (B, "\t\tS.%s[i].%s(B,L);}\n", F->name, M);
      break;
    case field_2darray:
      cbprintf (B, "\t{int i,j; for(i=0;i<%d;i++)\n", F->dim1);
      cbprintf (B, "\t\tfor(j=0;j<%d;j++)\n", F->dim2);
      cbprintf (B, "\t\t\tS.%s[i][j].%s(B,L);}\n", F->name, M);
      break;
    }
}

static void
serialize_raw (field_t * F, cbuffer_t * B)
{
  char *tname = 0;
  switch (F->type)
    {
    case typ_int:
      tname = "int";
      break;
    case typ_str:
      tname = "str";
      break;
    }
  switch (F->kind)
    {
    case field_single:
      cbprintf (B, "\tunparse_%s(B,L,V->%s);\n", tname, F->name);
      break;
    case field_array:
    case field_sarray:
      cbprintf (B, "\t{int i; for(i=0;i<%d;i++)\n", F->dim1);
      cbprintf (B, "\t\tunparse_%s(B,L,V->%s[i]);}\n", tname, F->name);
      break;
    case field_2darray:
      cbprintf (B, "\t{int i,j; for(i=0;i<%d;i++)\n", F->dim1);
      cbprintf (B, "\t\tfor(j=0;j<%d;j++)\n", F->dim2);
      cbprintf (B, "\t\t\tunparse_%s(B,L,obj->%s[i][j]);}\n", tname, F->name);
      break;
    }
}

static void
deserialize_raw (field_t * F, cbuffer_t * B)
{
  char *tname = 0;
  switch (F->type)
    {
    case typ_int:
      tname = "int";
      break;
    case typ_str:
      tname = "str";
      break;
    }
  switch (F->kind)
    {
    case field_single:
      cbprintf (B, "\tobj->%s=parse_%s(B,L);\n", F->name, tname);
      break;
    case field_array:
    case field_sarray:
      cbprintf (B, "\t{int i; for(i=0;i<%d;i++)\n", F->dim1);
      cbprintf (B, "\t\tobj->%s[i]=parse_%s(B,L);}\n", F->name, tname);
      break;
    case field_2darray:
      cbprintf (B, "\t{int i,j; for(i=0;i<%d;i++)\n", F->dim1);
      cbprintf (B, "\t\tfor(j=0;j<%d;j++)\n", F->dim2);
      cbprintf (B, "\t\t\tobj->%s[i][j]=parse_%s(B,L);}\n", F->name, tname);
      break;
    }
}
