/*

Copyright (c) 2009 Şükrü Çınar  (sukru cinar)

This program is distributed under GNU GPL v2. See file COPYING
for licensing details.

*/

#include <sys/types.h>
#include <regex.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "parsedoc.h"

#include "style.h"

static void apply_style (xcell_t * X, styleregion_t * R);
static void get_merge_coords (char *mc, int *sR, int *sC, int *eR, int *eC);
static void parse_cell_name (char *cn, int *row, int *col);
static int parse_col_name (char *x);
static char **strsplit (char *re, char *d);
static char *collect_substr (char *str, int len);
static void do_escape_sequences (char *X);

void *__unused_style_c[] = {

  0,
};

grid_t *
apply_styles (sheet_t * sheet)
{
  grid_t *G;
  int i, j;

  G = (typeof (G)) calloc (1, sizeof (G[0]));

  G->ncols = 0;
  G->nrows = 0;
  {
    typeof ((*sheet->cells).head.next) N9, T9;
    for (T9 = (*sheet->cells).head.next;
	 T9 != &((*sheet->cells).head); T9 = N9)
      {
	typeof (T9->data) C;
	C = T9->data;
	N9 = T9->next;

	if (C->row > G->nrows)
	  G->nrows = C->row;
	if (C->col > G->ncols)
	  G->ncols = C->col;

      }
  }

  G->nrows++;
  G->ncols++;

  G->cells = (typeof (G->cells)) calloc (G->nrows, sizeof (G->cells[0]));

  for (i = 0; i < G->nrows; i++)

    G->cells[i] =
      (typeof (G->cells[i])) calloc (G->ncols, sizeof (G->cells[i][0]));

  for (i = 0; i < G->nrows; i++)
    for (j = 0; j < G->ncols; j++)
      {
	G->cells[i][j].row = i;
	G->cells[i][j].col = j;
	G->cells[i][j].kind = ck_empty;
	G->cells[i][j].colspan = 1;
	G->cells[i][j].rowspan = 1;
	G->cells[i][j].bgcolor = 0xffFFff;
      }

  {
    typeof ((*sheet->styleregions).head.next) N13, T13;
    for (T13 = (*sheet->styleregions).head.next;
	 T13 != &((*sheet->styleregions).head); T13 = N13)
      {
	typeof (T13->data) K;
	K = T13->data;
	N13 = T13->next;

	for (i = K->startRow; i <= K->endRow; i++)
	  if (i >= 0 && i < G->nrows)
	    for (j = K->startCol; j <= K->endCol; j++)
	      if (j >= 0 && j < G->ncols)
		apply_style (&G->cells[i][j], K);

      }
  }

  {
    typeof ((*sheet->cells).head.next) N15, T15;
    for (T15 = (*sheet->cells).head.next;
	 T15 != &((*sheet->cells).head); T15 = N15)
      {
	typeof (T15->data) C;
	C = T15->data;
	N15 = T15->next;

	xcell_t *X;
	X = &G->cells[C->row][C->col];
	X->data = C->data;
	if (X->data)
	  do_escape_sequences (X->data);
	if (X->bold)
	  X->kind = ck_header;
	else
	  X->kind = ck_data;

      }
  }

  {
    typeof ((*sheet->merges).head.next) N17, T17;
    for (T17 = (*sheet->merges).head.next;
	 T17 != &((*sheet->merges).head); T17 = N17)
      {
	typeof (T17->data) R;
	R = T17->data;
	N17 = T17->next;

	int sR, sC, eR, eC;
	get_merge_coords (R->data, &sR, &sC, &eR, &eC);
	if (sR >= G->nrows || sC >= G->ncols)
	  continue;
	if (sR < 0)
	  sR = 0;
	if (sC < 0)
	  sC = 0;
	if (eR < sR)
	  continue;
	if (eC < sC)
	  continue;
	if (eR == sR && eC == sC)
	  continue;

	G->cells[sR][sC].rowspan = eR - sR + 1;
	G->cells[sR][sC].colspan = eC - sC + 1;
	for (i = sR; i <= eR; i++)
	  for (j = sC; j <= eC; j++)
	    {
	      if (i == sR && j == sC)
		continue;
	      G->cells[i][j].kind = ck_merged;
	      G->cells[sR][sC].border |= G->cells[i][j].border;
	      G->cells[i][j].merged_to = &G->cells[sR][sC];
	      G->cells[i][j].rowspan = 0;
	      G->cells[i][j].colspan = 0;
	    }

      }
  }

  return G;
}

static void
apply_style (xcell_t * X, styleregion_t * R)
{
  /* these magic numbers are retrieved by trying
     gnumeric 1.8.3. I wonder why they didnt provide
     short strings like center, right, etc. instead of numbers */
  switch (R->hAlign)
    {
    case 2:
      X->halign = al_begin;
      break;
    case 8:
      X->halign = al_center;
      break;
    case 4:
      X->halign = al_end;
      break;
    }
  switch (R->vAlign)
    {
    case 1:
      X->valign = al_begin;
      break;
    case 4:
      X->valign = al_center;
      break;
    case 2:
      X->valign = al_end;
      break;
    }
  if (R->bTop)
    X->border |= b_top;
  if (R->bBottom)
    X->border |= b_bottom;
  if (R->bLeft)
    X->border |= b_left;
  if (R->bRight)
    X->border |= b_right;
  if (R->bold)
    X->bold = 1;
  X->fgcolor = R->fgcolor;
  X->bgcolor = R->bgcolor;
}

static void get_merge_coords (char *mc, int *sR, int *sC, int *eR, int *eC)
{
  int i;
  char **comp;
  *sR = 0;
  *sC = 0;
  *eR = 0;
  *eC = 0;
  comp = strsplit (":", mc);
  if (!comp)
    return;
  if (comp[0])
    {
      parse_cell_name (comp[0], sR, sC);
      free (comp[0]);
    }
  else
    goto end;

  if (comp[1])
    {
      parse_cell_name (comp[1], eR, eC);
      free (comp[1]);
    }
  else
    goto end;

  for (i = 2; comp[i]; i++)
    free (comp[i]);
end:
  free (comp);
}

static void
parse_cell_name (char *cn, int *row, int *col)
{
  char c;
  int i;
  for (i = 0; cn[i] >= 'A' && cn[i] <= 'Z'; i++)
    ;
  if (cn[i])
    *row = atoi (cn + i) - 1;
  c = cn[i];
  cn[i] = 0;
  *col = parse_col_name (cn);
  cn[i] = c;
}

static int
parse_col_name (char *x)
{
  int S;
  S = 0;
  while (*x)
    {
      S = S * 26 + x[0] - 'A';
      x++;
    }
  return S;
}

static char **
strsplit (char *re, char *d)
{
  regex_t rx;
  int ec;
  if ((ec = regcomp (&rx, re, REG_EXTENDED)))
    {
      char ebuf[2000];
      regerror (ec, &rx, ebuf, sizeof (ebuf) - 1);
      exit (fprintf (stderr, "regcomp for '%s' failed: %s\n", re, ebuf));
    }
  char *orig_d = d;
  int nfields;
  char **fields = 0;

again:
  d = orig_d;
  nfields = 0;
  while (*d)
    {
      regmatch_t matches[2];
      if (regexec (&rx, d, 2, matches, 0))
	{
	  if (fields)
	    fields[nfields] = collect_substr (d, strlen (d));
	  nfields++;
	  d += strlen (d);
	}
      else
	{
	  if (matches[0].rm_so == 0)
	    {
	      d += matches[0].rm_eo;
	    }
	  else
	    {
	      if (fields)
		fields[nfields] = collect_substr (d, matches[0].rm_so);
	      d += matches[0].rm_eo;
	      nfields++;
	    }
	}
    }
  if (fields)
    {
      regfree (&rx);
      return fields;
    }
  if (!nfields)
    {
      regfree (&rx);
      return 0;
    }

  fields = (typeof (fields)) calloc ((nfields + 1), sizeof (fields[0]));

  fields[nfields] = 0;
  goto again;
}

static char *
collect_substr (char *str, int len)
{
  char *x;

  x = (typeof (x)) calloc ((len + 1), sizeof (x[0]));

  memcpy (x, str, len);
  x[len] = 0;
  return x;
}

static void
do_escape_sequences (char *X)
{
  int i, j;
  j = 0;
  for (i = 0; X[i]; i++)
    {
      if (X[i] != '!' && X[i] != '\\')
	{
	  X[j++] = X[i];
	  continue;
	}
      if (X[i] == '!')
	{
	  X[j++] = '\n';
	  continue;
	}
    }
  X[j] = 0;
}
