
namespace FX {
class FXAPI flight_log_table : public FXScrollArea {
	FXDECLARE(flight_log_table)
protected:
	flight_log_table();
private:
	flight_log_table(const flight_log_table&);
	flight_log_table &operator=(const flight_log_table&);

	FXApp* myapp;

private:
	int input_r;
	int input_c;

	FXFont *data_font;
	FXFont *header_font;
	int data_font_ascent;
	int header_font_ascent;
	int data_font_height;
	int header_font_height;
	int data_font_swidth;
	int header_font_swidth;

	FXTextField *editor;
private:
	typedef struct { int y; int x1; int x2; } xline_t;
	typedef struct { int x; int y1; int y2; } yline_t;
	typedef struct { int x1,y1; int x2,y2; } crect_t;
	typedef struct { crect_t area; unsigned color; } bgfill_def_t;
	typedef struct { crect_t area; FXColor color; } bgfill_act_t;

	typedef enum {  field_date,  field_atype,  field_areg,  field_depplc,  field_deptm,  field_arrplc,  field_arrtm,  field_se,  field_me,  field_mptime,  field_ftime,  field_ldg_day,  field_ldg_ngt,  field_opngt,  field_opifr,  field_ophood,  field_fpic,  field_fcop,  field_fdual,  field_finst,  field_synth,  field_rmk,  field_pg_mptime,  field_pg_ftime,  field_pg_ldg_day,  field_pg_ldg_ngt,  field_pg_opngt,  field_pg_opifr,  field_pg_ophood,  field_pg_fpic,  field_pg_fcop,  field_pg_fdual,  field_pg_finst,  field_pg_synth,  field_pp_mptime,  field_pp_ftime,  field_pp_ldg_day,  field_pp_ldg_ngt,  field_pp_opngt,  field_pp_opifr,  field_pp_ophood,  field_pp_fpic,  field_pp_fcop,  field_pp_fdual,  field_pp_finst,  field_pp_synth,  field_tot_mptime,  field_tot_ftime,  field_tot_ldg_day,  field_tot_ldg_ngt,  field_tot_opngt,  field_tot_opifr,  field_tot_ophood,  field_tot_fpic,  field_tot_fcop,  field_tot_fdual,  field_tot_finst,  field_tot_synth,   } cell_storage_kind_t;
	typedef enum { cell_int, cell_str } cell_base_type_t;

	typedef enum
		{ ck_merged, ck_empty, ck_data, ck_header }
			cell_kind_t;

	typedef struct
		{ cell_kind_t kind; int halign, valign;
		   int rowspan, colspan;
		   int merge_r, merge_c; int editable;
			cell_base_type_t  base_type;
		   cell_storage_kind_t skind;
		   int idx1, idx2;
		   const char *format;
		   int show_if_zero;
		 } cell_def_t;

	typedef struct
		{ int cwidth, cheight;
		char *data; } cell_t;

	static const xline_t horz_line_defs[23];
	static const yline_t vert_line_defs[24];
	static const bgfill_def_t bgfill_defs[31];
	static const cell_def_t cell_defs[21][23];

	xline_t horz_lines[23];
	yline_t vert_lines[24];
	bgfill_act_t bgfills[31];

	cell_t cells[21][23];

	int row_heights[21];
	int col_widths[23];
	int x_offsets[23+1];
	int y_offsets[21+1];

	int x_pad, y_pad;

	void get_text_size(int bold,char *data, int *w,int *h);
	void get_single_line_text_size
		(FXFont* font,char*data,int *tw,int *th);

	int render_text_line
		(FXDC* dc,FXFont *font,int X, int Y,
		int halign, char *data, int avail_w);
	void render_text
		(FXDC* dc,crect_t area, int valign,int halign, int bold,
		char *data,int content_width,int content_height);

	int calculate_col_widths();
	int calculate_row_heights();
	void calculate_cell_geometry(int r,int c);
	void calculate_cell_content_geometries();
	void convert_to_actual_coordinates();

	int ceil(int a,int b);
	int classify_point(crect_t R, int x,int y);
	int rectangles_intersect(crect_t R1,crect_t R2);
	int xline_visible(crect_t R, xline_t L);
	int yline_visible(crect_t R,yline_t L);
	FXColor rgb24tofxcolor(unsigned int x);
	int fit_to_range(int x,int min,int max);
	int unformat_int(char *txt,const char *fmt);
	char *unformat_str(char *txt,const char *fmt);
	char* format_int(int v,const char *fmt);
	char *format_str(char *txt,const char *fmt);

	void make_cell_visible(int r,int c);
public:
	class boxed_int {
	public:
		int value;
		flight_log_table* table;
		void set(int v);
		int row,col;
		void init(flight_log_table *tb,int r,int c);
		void start_edit();
		void serialize(char **buffer,size_t *len);
		void deserialize(char **buffer,size_t *len);
	};

	class boxed_str {
	public:
		char *value;
		flight_log_table* table;
		void set(char *v);
		int row,col;
		void init(flight_log_table *tb,int r,int c);
		void start_edit();
		void serialize(char **buffer,size_t *len);
		void deserialize(char **buffer,size_t *len);
	};

	struct {
			boxed_str date[16];
	boxed_str atype[16];
	boxed_str areg[16];
	boxed_str depplc[16];
	boxed_int deptm[16];
	boxed_str arrplc[16];
	boxed_int arrtm[16];
	boxed_str se[16];
	boxed_str me[16];
	boxed_int mptime[16];
	boxed_int ftime[16];
	boxed_int ldg_day[16];
	boxed_int ldg_ngt[16];
	boxed_int opngt[16];
	boxed_int opifr[16];
	boxed_int ophood[16];
	boxed_int fpic[16];
	boxed_int fcop[16];
	boxed_int fdual[16];
	boxed_int finst[16];
	boxed_int synth[16];
	boxed_str rmk[16];
	boxed_int pg_mptime;
	boxed_int pg_ftime;
	boxed_int pg_ldg_day;
	boxed_int pg_ldg_ngt;
	boxed_int pg_opngt;
	boxed_int pg_opifr;
	boxed_int pg_ophood;
	boxed_int pg_fpic;
	boxed_int pg_fcop;
	boxed_int pg_fdual;
	boxed_int pg_finst;
	boxed_int pg_synth;
	boxed_int pp_mptime;
	boxed_int pp_ftime;
	boxed_int pp_ldg_day;
	boxed_int pp_ldg_ngt;
	boxed_int pp_opngt;
	boxed_int pp_opifr;
	boxed_int pp_ophood;
	boxed_int pp_fpic;
	boxed_int pp_fcop;
	boxed_int pp_fdual;
	boxed_int pp_finst;
	boxed_int pp_synth;
	boxed_int tot_mptime;
	boxed_int tot_ftime;
	boxed_int tot_ldg_day;
	boxed_int tot_ldg_ngt;
	boxed_int tot_opngt;
	boxed_int tot_opifr;
	boxed_int tot_ophood;
	boxed_int tot_fpic;
	boxed_int tot_fcop;
	boxed_int tot_fdual;
	boxed_int tot_finst;
	boxed_int tot_synth;

	} S;

	typedef struct {
			char* date[16];
	char* atype[16];
	char* areg[16];
	char* depplc[16];
	int deptm[16];
	char* arrplc[16];
	int arrtm[16];
	char* se[16];
	char* me[16];
	int mptime[16];
	int ftime[16];
	int ldg_day[16];
	int ldg_ngt[16];
	int opngt[16];
	int opifr[16];
	int ophood[16];
	int fpic[16];
	int fcop[16];
	int fdual[16];
	int finst[16];
	int synth[16];
	char* rmk[16];
	int pg_mptime;
	int pg_ftime;
	int pg_ldg_day;
	int pg_ldg_ngt;
	int pg_opngt;
	int pg_opifr;
	int pg_ophood;
	int pg_fpic;
	int pg_fcop;
	int pg_fdual;
	int pg_finst;
	int pg_synth;
	int pp_mptime;
	int pp_ftime;
	int pp_ldg_day;
	int pp_ldg_ngt;
	int pp_opngt;
	int pp_opifr;
	int pp_ophood;
	int pp_fpic;
	int pp_fcop;
	int pp_fdual;
	int pp_finst;
	int pp_synth;
	int tot_mptime;
	int tot_ftime;
	int tot_ldg_day;
	int tot_ldg_ngt;
	int tot_opngt;
	int tot_opifr;
	int tot_ophood;
	int tot_fpic;
	int tot_fcop;
	int tot_fdual;
	int tot_finst;
	int tot_synth;

	} raw_storage_t;

		 void onccr_date(char *value,int idx);
	 void onccr_atype(char *value,int idx);
	 void onccr_areg(char *value,int idx);
	 void onccr_depplc(char *value,int idx);
	 void onccr_deptm(int value,int idx);
	 void onccr_arrplc(char *value,int idx);
	 void onccr_arrtm(int value,int idx);
	 void onccr_se(char *value,int idx);
	 void onccr_me(char *value,int idx);
	 void onccr_mptime(int value,int idx);
	 void onccr_ftime(int value,int idx);
	 void onccr_ldg_day(int value,int idx);
	 void onccr_ldg_ngt(int value,int idx);
	 void onccr_opngt(int value,int idx);
	 void onccr_opifr(int value,int idx);
	 void onccr_ophood(int value,int idx);
	 void onccr_fpic(int value,int idx);
	 void onccr_fcop(int value,int idx);
	 void onccr_fdual(int value,int idx);
	 void onccr_finst(int value,int idx);
	 void onccr_synth(int value,int idx);
	 void onccr_rmk(char *value,int idx);
	 void onccr_pg_mptime(int value);
	 void onccr_pg_ftime(int value);
	 void onccr_pg_ldg_day(int value);
	 void onccr_pg_ldg_ngt(int value);
	 void onccr_pg_opngt(int value);
	 void onccr_pg_opifr(int value);
	 void onccr_pg_ophood(int value);
	 void onccr_pg_fpic(int value);
	 void onccr_pg_fcop(int value);
	 void onccr_pg_fdual(int value);
	 void onccr_pg_finst(int value);
	 void onccr_pg_synth(int value);
	 void onccr_pp_mptime(int value);
	 void onccr_pp_ftime(int value);
	 void onccr_pp_ldg_day(int value);
	 void onccr_pp_ldg_ngt(int value);
	 void onccr_pp_opngt(int value);
	 void onccr_pp_opifr(int value);
	 void onccr_pp_ophood(int value);
	 void onccr_pp_fpic(int value);
	 void onccr_pp_fcop(int value);
	 void onccr_pp_fdual(int value);
	 void onccr_pp_finst(int value);
	 void onccr_pp_synth(int value);
	 void onccr_tot_mptime(int value);
	 void onccr_tot_ftime(int value);
	 void onccr_tot_ldg_day(int value);
	 void onccr_tot_ldg_ngt(int value);
	 void onccr_tot_opngt(int value);
	 void onccr_tot_opifr(int value);
	 void onccr_tot_ophood(int value);
	 void onccr_tot_fpic(int value);
	 void onccr_tot_fcop(int value);
	 void onccr_tot_fdual(int value);
	 void onccr_tot_finst(int value);
	 void onccr_tot_synth(int value);


	void serialize_data(char **buf,size_t *len);
	void deserialize_data(char *buf, size_t len);

	static raw_storage_t* parse_data(char *buf,size_t len);
	static void unparse_data(raw_storage_t *V,
		char **buf,size_t*len);
	static void destroy_data(raw_storage_t *obj);

	int show_if_zero(int row,int col);

public:
	flight_log_table(FXComposite* p,
		FXuint opts=FRAME_NORMAL,
		FXint x=0,FXint y=0,FXint w=0,FXint h=0);
	virtual ~flight_log_table();

	enum { ID_UPDATECELL= FXScrollBar::ID_LAST+1, };

	enum { valign_top, valign_center, valign_bottom };
	enum { halign_left, halign_center, halign_right };

	virtual bool canFocus() const { return true; }
	virtual void create();
	virtual void detach();

	virtual int getContentWidth();
	virtual int getContentHeight();
	virtual FXint getDefaultWidth();
	virtual FXint getDefaultHeight();

	void begin_update();
	void end_update();

	virtual void layout();

	long onPaint(FXObject*,FXSelector,void*);
	long onClick(FXObject*,FXSelector,void*);
	long onCellChange(FXObject*,FXSelector,void*);

public:
	int col(const char* X);
	int editable;
	void start_edit(int r,int c);
	void set_cell_text(int r,int c,char *text);
	void set_cell_text_noupdate(int r,int c,char *text);
};

}
